import json
import urllib2

from upd89.classes.encoder import MyEncoder
from upd89.lib.httpsclientauthconnection import HTTPSClientAuthConnection


def __getBasePath():
    return "/api/v2"


def getRegisterPath(_config):
    return __getBasePath() + "/register"


def getSystemNotifyPath(_config, urn):
    return __getBasePath() + "/system/" + urn + "/notify"


def getSystemNotifyHashPath(_config, urn):
    return __getBasePath() + "/system/" + urn + "/notify-hash"


def getSystemRefreshInstalledPath(_config, urn):
    return __getBasePath() + "/system/" + urn + "/refresh-installed"


def getSystemRefreshInstalledHashPath(_config, urn):
    return __getBasePath() + "/system/" + urn + "/refresh-installed-hash"


def getTaskNotifyPath(_config, taskid):
    return __getBasePath() + "/task/" + taskid + "/notify"


def push(_config, path, data):
    host = _config.getServerHost()
    port = _config.getServerPort()
    ca = _config.getTlsPath() + '/' + _config.getTlsCa()
    key = _config.getTlsPath() + '/' + _config.getTlsPrivKey()
    crt = _config.getTlsPath() + '/' + _config.getTlsPubCert()
    headers = {"Content-type": "application/json"}
    jsondata = json.dumps(data, cls=MyEncoder)
    answer = ''

    # print host, port, ca, key, crt, headers
    # print jsondata

    try:
        conn = HTTPSClientAuthConnection(
            host, port, key_file=key, cert_file=crt, ca_file=ca)
        conn.request('POST', path, jsondata, headers)
        response = conn.getresponse()
        print response.status, response.reason
        answer = response.read()
        conn.close()
    except Exception as e:
        print('exception!')
        print(type(e))
        print(e)

    return answer


def pushRegister(_config, sys):
    path = getRegisterPath(_config)
    return push(_config, path, sys)


def pushSystemNotifyHash(_config, urn, sys):
    path = getSystemNotifyHashPath(_config, urn)
    return push(_config, path, sys)


def pushSystemNotify(_config, urn, sys):
    path = getSystemNotifyPath(_config, urn)
    return push(_config, path, sys)


def pushSystemRefreshInstalledHash(_config, urn, packages):
    path = getSystemRefreshInstalledHashPath(_config, urn)
    return push(_config, path, packages)


def pushSystemRefreshInstalled(_config, urn, packages):
    path = getSystemRefreshInstalledPath(_config, urn)
    return push(_config, path, packages)


def pushTaskNotify(_config, taskid, tasknotify):
    path = getTaskNotifyPath(_config, taskid)
    return push(_config, path, tasknotify)



import shelve


class Persist:

    def __init__(self, filename):
        self.store = shelve.open(filename)

    def __contains__(self, key):
        return key in self.store

    def delete_key(self, key):
        if (key in self.store):
            del self.store[key]

    def get_key(self, key):
        if (key in self.store):
            return(self.store[key])
        else:
            return('')

    def get_keys(self):
        return self.store.keys()

    def set_key(self, key, value):
        self.store[key] = value

    def close(self):
        self.store.close()



class System:

    def __init__(self, name, urn, os, address, reboot_required):
        self.name = name
        self.urn = urn
        self.os = os
        self.address = address
        self.rebootRequired = reboot_required
        self.updCount = 0
        self.packageUpdates = list()

    def increment(self):
        self.updCount += 1

    def addUpdate(self, update):
        self.packageUpdates.append(update)
        self.increment()



class Package:

    def __init__(self, name, section, summary, homepage, installed_version,
                 is_base_version, base_version):
        self.name = name
        self.section = section
        self.summary = summary
        self.homepage = homepage
        self.installedVersion = installed_version
        self.isBaseVersion = is_base_version
        self.baseVersion = base_version



class PackageVersion:

    def __init__(self, version, sha256, arch, repository):
        self.version = version
        self.sha256 = sha256
        self.architecture = arch
        self.repository = repository



import os
import platform
import socket

from upd89.classes import system_register, system_notify


def get_hostname():
    return socket.gethostname()


def get_fqdn():
    return socket.getfqdn()


def get_urn():
    return 'demo-' + get_hostname() + '-demo'


def get_distro():
    return ' '.join(platform.linux_distribution())


def get_reboot_required():
    return os.path.isfile("/var/run/reboot-required")


def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 0))
        IP = s.getsockname()[0]
    except:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP

    # also possible:
    # my_ip = load(urlopen('http://jsonip.com'))['ip']


def get_register_system(port="8080"):
    myHostname = get_hostname()
    myURN = get_urn()
    myDistro = get_distro()
    myIP = get_ip()
    myAddress = get_fqdn() + ":" + port
    return system_register.System(name=myHostname, urn=myURN,
                                          os=myDistro, address=myAddress, tag="")


def get_notify_system(port="8080"):
    myHostname = get_hostname()
    myURN = get_urn()
    myDistro = get_distro()
    myIP = get_ip()
    myAddress = get_fqdn() + ":" + port
    needReboot = get_reboot_required()
    return system_notify.System(name=myHostname, urn=myURN,
                                        os=myDistro, address=myAddress,
                                        reboot_required=needReboot)



import socket
import ssl
import httplib

# http://code.activestate.com/recipes/577548-https-httplib-client-connection-with-certificate-v/


class HTTPSClientAuthConnection(httplib.HTTPSConnection):

    """ Class to make a HTTPS connection, with support for full client-based SSL Authentication"""

    def __init__(self, host, port, key_file, cert_file, ca_file, timeout=None):
        httplib.HTTPSConnection.__init__(
            self, host, key_file=key_file, cert_file=cert_file)
        self.key_file = key_file
        self.cert_file = cert_file
        self.ca_file = ca_file
        self.timeout = timeout

    def connect(self):
        """ Connect to a host on a given (SSL) port.
            If ca_file is pointing somewhere, use it to check Server Certificate.

            Redefined/copied and extended from httplib.py:1105 (Python 2.6.x).
            This is needed to pass cert_reqs=ssl.CERT_REQUIRED as parameter to ssl.wrap_socket(),
            which forces SSL to check server certificate against our client certificate.
        """
        sock = socket.create_connection((self.host, self.port), self.timeout)
        if self._tunnel_host:
            self.sock = sock
            self._tunnel()
        # force Server Certificate Check
        self.sock = ssl.wrap_socket(sock,
                                    keyfile=self.key_file,
                                    certfile=self.cert_file,
                                    ca_certs=self.ca_file,
                                    cert_reqs=ssl.CERT_REQUIRED)




#!/usr/bin/env python

import os
import sys
import signal
import argparse
from time import sleep
from daemonize import Daemonize
import schedule

from lib.configloader import ConfigLoader
from lib import mission
import lib.log


def _load_config(configfile):
    global _config, _log, _logger, _pid
    _config = ConfigLoader(configfile)
    _log = lib.log.Log(_config)
    _logger = _log.getLogger()
    _pid = _config.getPidFile()


def register():
    mission.send_register(_config, _logger)


def refreshinstalled():
    # mission.send_system_refreshinstalled(_config, _logger)
    mission.send_system_refreshinstalled_hash(_config, _logger)


def system_notify():
    mission.update_cache()
    # mission.send_system_notify(_config, _logger)
    mission.send_system_notify_hash(_config, _logger)


def do_update():
    #_logger.debug("Checking for new Tasks...")
    mission.do_update(_config, _logger)


def main():
    if not _config.is_registered():
        register()
    refreshinstalled()
    system_notify()
    schedule.every(2).hours.do(refreshinstalled)
    schedule.every(10).minutes.do(system_notify)
    schedule.every(30).seconds.do(do_update)

    while True:
        schedule.run_pending()
        sleep(5)


def start(configfile):
    _load_config(configfile)
    os.environ["UPD89_DATADIR"] = _config.getDataDir()

    newpid = os.fork()
    if newpid == 0:
        import websrv
        websrv.start(_config)
    else:
        daemon = Daemonize(app="test_app", pid=_pid, action=main,
                           keep_fds=_log.getKeepfds())
        daemon.start()


def stop(configfile):
    _load_config(configfile)

    import websrv
    websrv.stop(_config)

    with open(_pid, "r") as p:
        pid = int(p.read())
        os.kill(pid, signal.SIGTERM)


def debug(configfile):
    _load_config(configfile)
    main()




import configparser
import os.path


class ConfigLoader:

    def __init__(self, configname):
        if os.path.isfile(configname):
            self.config = configparser.ConfigParser()
            self.configname = configname
            self.config.read(self.configname)
        else:
            print("Configfile (%s) not found." % configname)
            exit(1)

    def __str__(self):
        return repr(self)

    def getWebserverPort(self):
        return(self.config['webserver']['port'])

    def getWebserverLogfile(self):
        return(self.config['webserver']['logfile'])

    def getWebserverPidfile(self):
        return(self.config['webserver']['pidfile'])

    def getServerHost(self):
        return(self.config['controlcenter']['host'])

    def getServerPort(self):
        return(self.config['controlcenter']['port'])

    def getTlsPath(self):
        return(self.config['tls']['path'])

    def getTlsCa(self):
        return(self.config['tls']['cacert'])

    def getTlsPrivKey(self):
        return(self.config['tls']['privkey'])

    def getTlsPubCert(self):
        return(self.config['tls']['pubcert'])

    def getDataDir(self):
        return(self.config['agent']['datadir'])

    def getLogFile(self):
        return(self.config['agent']['logfile'])

    def getPidFile(self):
        return(self.config['agent']['pidfile'])

    def is_registered(self):
        return self.__ret_if_exists('controlcenter', 'registered') == "true"

    def set_registered(self):
        self.__set_config('controlcenter', 'registered', "true")

    def reloadConfig(self):
        self.config.read(self.configname)

    def __set_config(self, group, tag, value):
        self.config[group][tag] = value
        with open(self.configname, 'w') as configfile:
            self.config.write(configfile)

    def __ret_if_exists(self, group, tag):
        if self.config.has_option(group, tag):
            return(self.config[group][tag])
        else:
            return None



import logging


class screenLog:

    def debug(self, msg):
        print(msg)


class Log:

    def __init__(self, _config):
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.DEBUG)
        self.logger.propagate = False
        self.fh = logging.FileHandler(_config.getLogFile(), "w")
        self.fh.setLevel(logging.DEBUG)
        self.logger.addHandler(self.fh)
        self.keep_fds = [self.fh.stream.fileno()]

    def getLogger(self):
        return self.logger

    def getKeepfds(self):
        return self.keep_fds



class Packagelist:

    def __init__(self):
        self.pkgCount = 0
        self.packages = list()

    def increment(self):
        self.pkgCount += 1

    def add(self, package):
        self.packages.append(package)
        self.increment()



#!/usr/bin/env python

from lib.bottletlsdaemon import daemon_run
from bottle import get, post, request, run, abort, redirect
import os
import json
import lib.persist


@get('/')
def index():
    return "nothing to see here"


@post('/task')
def new_task():
    cwd = os.getenv('UPD89_DATADIR', "/etc/upd89")
    tasks = lib.persist.Persist(cwd + "/tasks.data")
    data = json.load(request.body)
    tasks.set_key(data.get('task_id').encode("utf8"), request.body.read())
    tasks.close()
    return("ok")


def start(config=None):
    if config == None:
        daemon_run(
            host     = '0.0.0.0',
            port     = os.environ["UPD89_WEBSRV_PORT"],
            pidfile  = os.environ["UPD89_WEBSRV_PIDFILE"],
            logfile  = os.environ["UPD89_WEBSRV_LOGFILE"],
            keyfile  = os.environ["UPD89_WEBSRV_KEYFILE"],
            certfile = os.environ["UPD89_WEBSRV_CERTFILE"],
            cafile   = os.environ["UPD89_WEBSRV_CAFILE"],
            action   = "start"
        )
    else:
        cpath = config.getTlsPath()
        daemon_run(
            host     = '0.0.0.0',
            port     = config.getWebserverPort(),
            pidfile  = config.getWebserverPidfile(),
            logfile  = config.getWebserverLogfile(),
            keyfile  = cpath + '/' + config.getTlsPrivKey(),
            certfile = cpath + '/' + config.getTlsPubCert(),
            cafile   = cpath + '/' + config.getTlsCa(),
            action   = "start"
        )


def stop(config=None):
    if config == None:
        daemon_run(
            pidfile = os.environ["UPD89_WEBSRV_PIDFILE"],
            action  = "stop"
        )
    else:
        daemon_run(
            pidfile = config.getWebserverPidfile(),
            action  = "stop"
        )





import json
import datetime
from upd89.lib import persist, upstream, sysinfo, pkg


def update_cache():
    pkg.updateCache()


def send_register(_config, _logger):
    sys = sysinfo.get_register_system()
    _logger.debug("Sending to server (register %s)..." %
                  sysinfo.get_hostname())
    response = upstream.pushRegister(_config, sys)
    _logger.debug("Response: " + response)
    _config.set_registered()


def send_system_refreshinstalled_hash(_config, _logger):
    known_packages = persist.Persist(_config.getDataDir() + "/known_packages.data")
    packages = pkg.getPackageHashList(known_packages.get_keys())
    _logger.debug("Sending to server (refreshInstalledHash %s)..." %
                  sysinfo.get_hostname())
    response = upstream.pushSystemRefreshInstalledHash(_config, sysinfo.get_urn(), packages)
    _logger.debug("Response: " + response)

    j = json.loads(response)
    print j.get("status")
    for h in j.get("knownPackages"):
        now = datetime.datetime.now().strftime("%d.%m.%y %H:%M")
        known_packages.set_key(h.encode('ascii', 'ignore'), now)
        packages.packages.remove(h)
    print("unknown packages: %d" % len(packages.packages))
    known_packages.close() # update 'known packages' after this?

    # sending full information for unknown packages
    if (len(packages.packages) > 0):
        complete_packages = pkg.getPackageListIncremental(packages.packages)
        _logger.debug("Sending to server (refreshInstalled %s)..." %
                      sysinfo.get_hostname())
        response = upstream.pushSystemRefreshInstalled(_config, sysinfo.get_urn(), complete_packages)
        _logger.debug("Response: " + response)
        j = json.loads(response)

    # sending full list if we had a count mismatch
    if j.get("status") == "countMismatch":
        packages = pkg.getPackageHashList(list())
        _logger.debug("Sending to server (refreshInstalledHash after countMismatch %s)..." %
                       sysinfo.get_hostname())
        response = upstream.pushSystemRefreshInstalledHash(_config, sysinfo.get_urn(), packages)
        _logger.debug("Response: " + response)
        j = json.loads(response)
        print(j.get("status"))


def send_system_refreshinstalled(_config, _logger):
    packages = pkg.getPackageList()
    _logger.debug("Sending to server (refreshInstalled %s)..." %
                  sysinfo.get_hostname())
    response = upstream.pushSystemRefreshInstalled(_config, sysinfo.get_urn(), packages)
    _logger.debug("Response: " + response)


def send_system_notify_hash(_config, _logger):
    known_updates = persist.Persist(_config.getDataDir() + "/known_updates.data")
    sys = sysinfo.get_notify_system()
    sys = pkg.addUpdateHashes(sys, known_updates.get_keys())
    _logger.debug("Sending to server (notifyHash %s)..." %
                  sysinfo.get_hostname())
    response = upstream.pushSystemNotifyHash(_config, sysinfo.get_urn(), sys)
    _logger.debug("Response: " + response)

    j = json.loads(response)
    print j.get("status")
    for h in j.get("knownPackages"):
        now = datetime.datetime.now().strftime("%d.%m.%y %H:%M")
        known_updates.set_key(h.encode('ascii', 'ignore'), now)
        sys.packageUpdates.remove(h)
    print("unknown updates: %d" % len(sys.packageUpdates))
    known_updates.close() # update 'known updates' after this?

    # sending full information for unknown updates
    if (len(sys.packageUpdates) > 0):
        complete_sys = sysinfo.get_notify_system()
        complete_sys = pkg.addUpdatesIncremental(complete_sys, sys.packageUpdates)
        _logger.debug("Sending to server (notify %s)..." %
                      sysinfo.get_hostname())
        response = upstream.pushSystemNotify(_config, sysinfo.get_urn(), complete_sys)
        _logger.debug("Response: " + response)
        j = json.loads(response)

    # sending full list if we had a count mismatch
    if j.get("status") == "countMismatch":
        sys = sysinfo.get_notify_system()
        sys = pkg.addUpdateHashes(sys, list())
        _logger.debug("Sending to server (notifyHash after countMismatch %s)..." %
                      sysinfo.get_hostname())
        response = upstream.pushSystemNotifyHash(_config, sysinfo.get_urn(), sys)
        _logger.debug("Response: " + response)
        j = json.loads(response)
        print(j.get("status"))


def send_system_notify(_config, _logger):
    sys = sysinfo.get_notify_system()
    sys = pkg.addUpdates(sys)
    _logger.debug("Sending to server (notify %s)..." % sysinfo.get_hostname())
    response = upstream.pushSystemNotify(_config, sysinfo.get_urn(), sys)
    _logger.debug("Response: " + response)


def do_update(_config, _logger):
    tasks = persist.Persist(_config.getDataDir() + "/tasks.data")
    for key in tasks.get_keys():
        json_data = tasks.get_key(key)
        _logger.debug(u"key: %s - json: %s" % (key, json_data))
        t = json.loads(json_data)
        p_list = list()
        for p in t.get("packages"):
            pkg_name = p.get("pkg_name")
            pkg_version = p.get("pdk_version")
            p_list.append(pkg_name)
        tasknotify = pkg.do_update(p_list)
        response = upstream.pushTaskNotify(_config, key, tasknotify)
        _logger.debug("Response: " + response)
        tasks.delete_key(key)
    tasks.close()



#!/usr/bin/env python

import os
from setuptools import setup

setup(
    name = "upd89",
    version = "0.5.26",
    author = "Ueli Bosshard",
    author_email = "ubosshard@gmail.com",
    description = ("upd89 is a system update management "
                   "for debian based systems."),
    license = "MIT",
    keywords = "upd89 system update orchestration",
    url = "http://upd89.org",
    packages = ['upd89', 'upd89.classes', 'upd89.lib'],
    scripts=['bin/upd89'],
    data_files=[
        ('/etc/upd89', ['etc/config.ini']),
        ('/etc/init.d', ['init-script/upd89']),
    ],
    long_description='upd89 is a system update management ' +
        'for debian based systems.',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
    install_requires=[
        'daemonize',
        'schedule',
        'configparser',
        'bottle',
    ],
)




class Update:

    def __init__(self, name, candidate_version, baseversion_hash):
        self.name = name
        self.candidateVersion = candidate_version
        self.baseVersionHash = baseversion_hash



import os
import ssl
import argparse
import signal
import daemon
import lockfile
import bottle
from contextlib import contextmanager

# https://raw.githubusercontent.com/jonathanhood/bottledaemon/master/bottledaemon/bottledaemon.py
# http://www.socouldanyone.com/2014/01/bottle-with-ssl.html


class SSLWSGIRefServer(bottle.ServerAdapter):
    def __init__(self, host='128.0.0.1', port=8080, keyfile='priv.key',
                 certfile='pub.crt', cafile='ca.crt', **options):
        #super(SSLWSGIRefServer, self).__init__(host, port, options)
        self.keyfile = keyfile
        self.certfile = certfile
        self.cafile = cafile
        self.options = options
        self.host = host
        self.port = int(port)

    def run(self, handler):
        from wsgiref.simple_server import make_server, WSGIRequestHandler
        import ssl
        if self.quiet:
            class QuietHandler(WSGIRequestHandler):

                def log_request(*args, **kw):
                    pass
            self.options['handler_class'] = QuietHandler
        srv = make_server(self.host, self.port, handler, **self.options)
        srv.socket = ssl.wrap_socket(
            srv.socket,
            keyfile=self.keyfile,
            certfile=self.certfile,
            ca_certs=self.cafile,
            cert_reqs=ssl.CERT_REQUIRED,
            server_side=True)
        srv.serve_forever()


@contextmanager
def __locked_pidfile(filename):
    # Acquire the lockfile
    lock = lockfile.FileLock(filename)
    lock.acquire(-1)

    # Write out our pid
    realfile = open(filename, "w")
    realfile.write(str(os.getpid()))
    realfile.close()

    # Yield to the daemon
    yield

    # Cleanup after ourselves
    os.remove(filename)
    lock.release()


def daemon_run(host="localhost", port="8080", pidfile=None, logfile=None,
               keyfile='priv.key', certfile='pub.crt', cafile='ca.crt',
               action="start"):
    """
    Get the bottle 'run' function running in the background as a daemonized
    process.

    :host: The host interface to listen for connections on. Enter 0.0.0.0
           to listen on all interfaces. Defaults to localhost.
    :port: The host port to listen on. Defaults to 8080.
    :pidfile: The file to use as the process id file. Defaults to "bottle.pid"
    :logfile: The file to log stdout and stderr from bottle to. Defaults to "bottle.log"

    """

    if pidfile is None:
        pidfile = os.path.join(
            os.getcwd(),
            "bottle.pid"
        )

    if logfile is None:
        logfile = os.path.join(
            os.getcwd(),
            "bottle.log"
        )

    if action == "start":
        log = open(logfile, "w+")
        context = daemon.DaemonContext(
            pidfile=__locked_pidfile(pidfile),
            stdout=log,
            stderr=log
        )

        with context:
            # bottle.run(host=host, port=port)
            srv = SSLWSGIRefServer(host=host, port=port, keyfile=keyfile,
                                   certfile=certfile, cafile=cafile)
            bottle.run(server=srv)
    else:
        with open(pidfile, "r") as p:
            pid = int(p.read())
            os.kill(pid, signal.SIGTERM)



import os
os.putenv("DEBIAN_FRONTEND", "noninteractive")
import sys
import apt
import apt_pkg
from upd89.classes.update import Update
from upd89.classes.package import Package
from upd89.classes.packagelist import Packagelist
from upd89.classes.repository import Repository
from upd89.classes.package_version import PackageVersion
from upd89.classes.task_notify import TaskNotify


def _getCache():
    return apt.Cache(apt.progress.text.OpProgress())

def _getBaseSha256(pkg):
    pkg_base = pkg.versions[-1]
    if (pkg_base.sha256 == ''):
        pkg_arch = pkg.architecture()
        return "no_hash_for_%s_%s_%s" % (pkg.name, pkg_arch, pkg_base.version)
    else:
        return pkg_base.sha256

def _getInstalledSha256(pkg):
    if (pkg.installed.sha256 == ''):
        pkg_arch = pkg.architecture()
        return "no_hash_for_%s_%s_%s" % (pkg.name, pkg_arch,
                                         pkg.installed.version)
    else:
        return pkg.installed.sha256

def _getCandidateSha256(pkg):
    if (pkg.candidate.sha256 == ''):
        pkg_arch = pkg.architecture()
        return "no_hash_for_%s_%s_%s" % (pkg.name, pkg_arch,
                                         pkg.candidate.version)
    else:
        return pkg.candidate.sha256

def _buildPackage(pkg):
    pkg_origin = pkg.installed.origins[0]
    pkg_base = pkg.versions[-1]
    pkg_base_origin = pkg_base.origins[0]
    pkg_arch = pkg.architecture()

    inst_sha256 = _getInstalledSha256(pkg)
    base_sha256 = _getBaseSha256(pkg)

    installed_repo = Repository(archive=pkg_origin.archive,
                                origin=pkg_origin.origin,
                                component=pkg_origin.component)

    installed_version = PackageVersion(version=pkg.installed.version,
                                       sha256=inst_sha256,
                                       arch=pkg_arch,
                                       repository=installed_repo)


    if (base_sha256 == inst_sha256):
        is_base_version = True
        base_version = False
    else:
        is_base_version = False
        base_repo = Repository(archive=pkg_base_origin.archive,
                               origin=pkg_base_origin.origin,
                               component=pkg_base_origin.component)

        base_version = PackageVersion(version=pkg_base.version,
                                      sha256=base_sha256,
                                      arch=pkg_arch,
                                      repository=base_repo)

    package = Package(name=pkg.name, section=pkg.section,
                      summary=pkg.installed.summary,
                      homepage=pkg.installed.homepage,
                      installed_version=installed_version,
                      is_base_version=is_base_version,
                      base_version=base_version)

    return package


def _buildUpdate(pkg):
    pkg_base = pkg.versions[-1]
    pkg_origin = pkg.versions[0].origins[0]
    pkg_arch = pkg.architecture()

    upd_sha256 = _getCandidateSha256(pkg)
    base_sha256 = _getBaseSha256(pkg)

    candidate_repo = Repository(archive=pkg_origin.archive,
                                origin=pkg_origin.origin,
                                component=pkg_origin.component)

    candidate_version = PackageVersion(version=pkg.candidate.version,
                                       sha256=upd_sha256,
                                       arch=pkg_arch,
                                       repository=candidate_repo)

    update = Update(name=pkg.name,
                    candidate_version=candidate_version,
                    baseversion_hash=base_sha256)

    return update


def updateCache():
    cache = _getCache()
    if os.geteuid() == 0:
        print("Updating cache...")
        try:
            cache.update()
        except:
            print("had some troubles, but anyway, let's go on...")
    cache.close()


def addUpdateHashes(sys, knownHashes):
    print("Reading local cache...")
    cache = _getCache()
    print("Reading Upgradable Packages...")
    for pkg in cache:
        if (pkg.is_upgradable):
            upd_sha256 = _getCandidateSha256(pkg)
            if not upd_sha256 in knownHashes:
                sys.addUpdate(upd_sha256)
            else:
                sys.increment()
    cache.close()
    return sys


def addUpdates(sys):
    print("Reading local cache...")
    cache = _getCache()
    print("Reading Upgradable Packages...")
    for pkg in cache:
        if (pkg.is_upgradable):
            sys.addUpdate(_buildUpdate(pkg))
    cache.close()
    return sys


def addUpdatesIncremental(sys, requestedPackages):
    print("Reading local cache...")
    cache = _getCache()
    print("Reading Upgradable Packages...")
    for pkg in cache:
        if (pkg.is_upgradable):
            upd_sha256 = _getCandidateSha256(pkg)
            if (upd_sha256 in requestedPackages):
                sys.addUpdate(_buildUpdate(pkg))
            else:
                sys.increment()
    cache.close()
    return sys


def getPackageHashList(knownHashes):
    cache = _getCache()
    packages = Packagelist()
    for pkg in cache:
        if (pkg.is_installed):
            inst_sha256 = _getInstalledSha256(pkg)
            if not inst_sha256 in knownHashes:
                packages.add(inst_sha256)
            else:
                packages.increment()
    cache.close()
    return packages


def getPackageList():
    cache = _getCache()
    packages = Packagelist()
    for pkg in cache:
        if (pkg.is_installed):
            packages.add(_buildPackage(pkg))
    cache.close()
    return packages


def getPackageListIncremental(requestedPackages):
    cache = _getCache()
    packages = Packagelist()
    for pkg in cache:
        if (pkg.is_installed):
            inst_sha256 = _getInstalledSha256(pkg)
            if (inst_sha256 in requestedPackages):
                packages.add(_buildPackage(pkg))
            else:
                packages.increment()
    cache.close()
    return packages


def do_update(p_list):

    class TextInstallProgress(apt.progress.base.InstallProgress):

        def __init__(self, progress_log):
            self.progress_log = progress_log
            super(TextInstallProgress, self).__init__()

        def fork(self):
            pid = os.fork()
            if pid == 0:
                os.dup2(self.progress_log, 1)
                os.dup2(self.progress_log, 2)
            return pid

    apt_pkg.config.set("APT::Get::Simulate", "true")
    apt_pkg.config.set("dir::cache", "/tmp")
    cache = _getCache()
    state = "Failed"
    progress_output, progress_input = os.pipe()

    for p in p_list:
        if p in cache:
            cache[p].mark_upgrade()
        else:
            print("package not found: '%s'" % p)

    try:
        result = cache.commit(apt.progress.text.AcquireProgress(),
                              TextInstallProgress(progress_input))
        if result:
            state = "Done"
    except SystemError as e:
        print("Error while updating: '%s'" % e)
    cache.close()
    os.close(progress_input)
    log = os.fdopen(progress_output).read()
    print("---\n" + log + "\n---")
    return(TaskNotify(state=state, log=log))
